-----------------------------------
-- Setting up scope and libs
-----------------------------------

local AddonName = "iGuild"
iGuild = LibStub("AceAddon-3.0"):NewAddon(AddonName, "AceEvent-3.0", "AceTimer-3.0");

local QTip = LibStub("LibQTip-1.0");
local Tourist = LibStub("LibTourist-3.0");
local L = LibStub("AceLocale-3.0"):GetLocale(AddonName);

local Crayon = LibStub("LibCrayon-3.0");

-----------------------------------------
-- Variables, functions and colors
-----------------------------------------

local Tooltip; -- our tooltip

local Roster = {}; -- the roster table, indexes are defined by the ROSTER_ values beneath
local RosterTimer; -- timer when the roster is fetched again

local ROSTER_CHAR_NAME, ROSTER_CHAR_LEVEL, ROSTER_CHAR_CLASS, ROSTER_CHAR_ZONE, ROSTER_CHAR_STATUS, ROSTER_CHAR_MOBILE = 1,2,3,4,5,6;
local ROSTER_ACM_POINTS, ROSTER_ACM_RANK, ROSTER_GUILD_RANK, ROSTER_GUILD_NOTE, ROSTER_OFFICER_NOTE, ROSTER_GUILD_RANK_NUM = 7,8,9,10,11,12;
local ROSTER_TRADESKILLS, ROSTER_CHAR_CLASS_LOCALIZED, ROSTER_XP_MAX = 13,14,15;

local TradeSkillDB; -- table for the tradeskill database, f.e. mining, skinning, etc.
local TradeSkillMates = {}; -- table[character] => { tradeskills }

local TRADE_ID, TRADE_ICON, TRADE_COLLAPSED = 1,2,3;

local MAX_LEVEL = 85; -- max level variable, dirty
local MAX_ACMPOINTS = 15000; -- dirty, too lazy to think about a method to calculate the maximum acm points

local COLOR_GOLD = "|cfffed100";
local COLOR_MASTER  = "|cffff6644";
local COLOR_OFFICER = "|cff40c040";

local function tclear(t, wipe)
	if( type(t) ~= "table" ) then return end;
	for k in pairs(t) do
		t[k] = nil;
	end
	t[''] = 1;
	t[''] = nil;
	if( wipe ) then
		t = nil;
	end
end

-----------------------------
-- Setting up the feed
-----------------------------

local function Feed_OnClick(_, button)
	if( button == "LeftButton" ) then
		if( IsAltKeyDown() and CanGuildInvite() ) then
			if( StaticPopup_FindVisible("ADD_GUILDMEMBER") ) then
				StaticPopup_Hide("ADD_GUILDMEMBER");
			else
				StaticPopup_Show("ADD_GUILDMEMBER");
			end
		else
			ToggleGuildFrame(1);
		end
	elseif( button == "RightButton" ) then
		iGuild:OpenOptions();
	end
end

local function Feed_OnEnter(anchor)
	if( not IsInGuild() ) then
		return;
	end
	
	for k, v in QTip:IterateTooltips() do
		if( type(k) == "string" and strsub(k, 1, 6) == "iSuite" ) then
			v:Release(k);
		end
	end
		
	Tooltip = QTip:Acquire("iSuite"..AddonName);
	Tooltip:SetAutoHideDelay(0.1, anchor);
	Tooltip:SmartAnchorTo(anchor);
	iGuild:UpdateTooltip();
	Tooltip:Show();
end

iGuild.Feed = LibStub("LibDataBroker-1.1"):NewDataObject(AddonName, {
	type = "data source",
	text = "",
	icon = "Interface\\Addons\\iGuild\\Images\\iGuild",
	OnClick = Feed_OnClick,
	OnEnter = Feed_OnEnter,
});

-----------------------------
-- Sorting and Columns
-----------------------------

iGuild.Sort = {
	name = function(a, b)
		return a[ROSTER_CHAR_NAME] < b[ROSTER_CHAR_NAME];
	end,
	level = function(a, b)
		if( a[ROSTER_CHAR_LEVEL] < b[ROSTER_CHAR_LEVEL] ) then
			return true;
		elseif( a[ROSTER_CHAR_LEVEL] > b[ROSTER_CHAR_LEVEL] ) then
			return false;
		else
			return iGuild.Sort.name(a, b);
		end
	end,
	class = function(a, b)
		if( a[ROSTER_CHAR_CLASS] < b[ROSTER_CHAR_CLASS] ) then
			return true;
		elseif( a[ROSTER_CHAR_CLASS] > b[ROSTER_CHAR_CLASS] ) then
			return false;
		else
			return iGuild.Sort.name(a, b);
		end
	end,
	rank = function(a, b)
		if( a[ROSTER_GUILD_RANK_NUM] < b[ROSTER_GUILD_RANK_NUM] ) then
			return true;
	elseif( a[ROSTER_GUILD_RANK_NUM] > b[ROSTER_GUILD_RANK_NUM] ) then
			return false;
		else
			return iGuild.Sort.name(a, b);
		end
	end,
	points = function(a, b)
		if( a[ROSTER_ACM_POINTS] > b[ROSTER_ACM_POINTS] ) then
			return true;
		elseif( a[ROSTER_ACM_POINTS] < b[ROSTER_ACM_POINTS] ) then
			return false;
		else
			return iGuild.Sort.name(a, b);
		end
	end,
	zone = function(a, b)
		if( a[ROSTER_CHAR_ZONE] < b[ROSTER_CHAR_ZONE] ) then
			return true;
		elseif( a[ROSTER_CHAR_ZONE] > b[ROSTER_CHAR_ZONE] ) then
			return false;
		else
			return iGuild.Sort.name(a, b);
		end
	end
};

iGuild.Columns = {
	level = {
		Label = L["Level"],
		Brush = function(v)
			local option = iGuild.db.Column.level.Color;
			if( option == 2 ) then -- encolor by difficulty
				local c = GetQuestDifficultyColor(v[ROSTER_CHAR_LEVEL]);
				return ("|cff%02x%02x%02x%s|r"):format(c.r *255, c.g *255, c.b *255, v[ROSTER_CHAR_LEVEL]);
			elseif( option == 3 ) then -- encolor by threshold
				return ("|cff%s%s|r"):format(Crayon:GetThresholdHexColor(v[ROSTER_CHAR_LEVEL], MAX_LEVEL), v[ROSTER_CHAR_LEVEL]);
			else
				return ("%s%s|r"):format(COLOR_GOLD, v[ROSTER_CHAR_LEVEL]);
			end
		end,
	},
	name = {
		Label = L["Name"],
		Brush = function(v)
			local status;
			if( v[ROSTER_CHAR_STATUS] == 1 ) then
				status = ("<%s>"):format(AFK);
			elseif( v[ROSTER_CHAR_STATUS] == 2 ) then
				status = ("<%s>"):format(DND);
			else
				status = "";
			end
			
			local option = iGuild.db.Column.name.Color;
			if( option == 2 ) then
				local c = RAID_CLASS_COLORS[v[ROSTER_CHAR_CLASS]];
				return ("|cff%02x%02x%02x%s%s|r"):format(c.r *255, c.g *255, c.b *255, status, v[ROSTER_CHAR_NAME]);
			else
				return ("%s%s%s|r"):format(COLOR_GOLD, status, v[ROSTER_CHAR_NAME]);
			end
		end,
	},
	zone = {
		Label = L["Zone"],
		Brush = function(v)
			local r, g, b = Tourist:GetFactionColor(v[ROSTER_CHAR_ZONE]);
			return ("|cff%02x%02x%02x%s|r"):format(r *255, g *255, b *255, v[ROSTER_CHAR_ZONE]);
		end,
	},
	rank = {
		Label = L["Rank"],
		Brush = function(v)
			local option = iGuild.db.Column.rank.Color;
			if( option == 2 ) then --encolor by threshold
				local maxr = GuildControlGetNumRanks();
				return ("|cff%s%s|r"):format(Crayon:GetThresholdHexColor(maxr - v[ROSTER_GUILD_RANK_NUM], maxr -1), v[ROSTER_GUILD_RANK]);
			else
				return ("%s%s|r"):format(COLOR_GOLD, v[ROSTER_GUILD_RANK]);
			end
		end,
		Script = function(_, v, button)
			if( button == "LeftButton" and CanGuildPromote() ) then
				GuildPromote(v[ROSTER_CHAR_NAME]);
			end
			if( button == "RightButton" and CanGuildDemote() ) then
				GuildDemote(v[ROSTER_CHAR_NAME]);
			end
		end,
		ScriptUse = function() return ( CanGuildPromote() or CanGuildDemote() ) end,
	},
	note = {
		Label = L["Note"],
		Brush = function(v)
			return ("%s%s|r"):format(COLOR_GOLD, v[ROSTER_GUILD_NOTE]);
		end,
	},
	officernote = {
		Label = L["OfficerNote"],
		Brush = function(v)
			return ("%s%s|r"):format(COLOR_OFFICER, v[ROSTER_OFFICER_NOTE]);
		end,
		CanUse = function() return CanViewOfficerNote() end,
	},
	notecombi = {
		Label = L["Note"],
		Brush = function(v)
			local normal = false;
			local officer = false;
			
			if( v[ROSTER_GUILD_NOTE] and v[ROSTER_GUILD_NOTE] ~= "" ) then
				normal = true;
			end
			if( CanViewOfficerNote() and v[ROSTER_OFFICER_NOTE] and v[ROSTER_OFFICER_NOTE] ~= "" ) then
				officer = true;
			end
			
			local note = "";
			
			if( normal and not officer ) then
				note = ("%s%s|r"):format(COLOR_GOLD, v[ROSTER_GUILD_NOTE]);
			elseif( officer and not normal ) then
				note = ("%s%s|r"):format(COLOR_OFFICER, v[ROSTER_OFFICER_NOTE]);
			elseif( normal and officer ) then
				note = ("%s%s|r / %s%s|r"):format(COLOR_GOLD, v[ROSTER_GUILD_NOTE], COLOR_OFFICER, v[ROSTER_OFFICER_NOTE]);
			end
			
			return note;
		end,
	},
	acmpoints = {
		Label = L["Points"],
		Brush = function(v)
			local option = iGuild.db.Column.acmpoints.Color;
			if( option == 2 ) then -- encolor by threshold
				return ("|cff%s%s|r"):format(Crayon:GetThresholdHexColor(v[ROSTER_ACM_POINTS], MAX_ACMPOINTS), v[ROSTER_ACM_POINTS]);
			else
				return ("%s%s|r"):format(COLOR_GOLD, v[ROSTER_ACM_POINTS]);
			end
		end,
	},
	tradeskills = {
		Label = L["TradeSkills"],
		Brush = function(v)
			local option = iGuild.db.Column.tradeskills.Enable;
			if( not option ) then
				return ("%s%s|r"):format(COLOR_GOLD, UNKNOWN);
			end
			
			local ts = TradeSkillMates[v[ROSTER_CHAR_NAME]];
			local label = "";

			if( ts ) then
				if( #ts >= 1 ) then
					label = ("|T%s:14:14|t"):format("Interface\\Addons\\iGuild\\Images\\"..iGuild:GetTradeSkill(ts[1], TRADE_ICON));
				end
				if( #ts >= 2 ) then
					label = ("%s |T%s:14:14|t"):format(label, "Interface\\Addons\\iGuild\\Images\\"..iGuild:GetTradeSkill(ts[2], TRADE_ICON));
				end
			end
			
			return label;
		end,
		Script = function(_, v, button)
			local ts = TradeSkillMates[v[ROSTER_CHAR_NAME]];
			if( not ts ) then
				return;
			end
			
			if( button == "LeftButton" and #ts >= 1 and CanViewGuildRecipes(iGuild:GetTradeSkill(ts[1], TRADE_ID)) ) then
				GetGuildMemberRecipes(v[ROSTER_CHAR_NAME], iGuild:GetTradeSkill(ts[1], TRADE_ID));
			elseif( button == "RightButton" and #ts >= 2 and CanViewGuildRecipes(iGuild:GetTradeSkill(ts[2], TRADE_ID)) ) then
				GetGuildMemberRecipes(v[ROSTER_CHAR_NAME], iGuild:GetTradeSkill(ts[2], TRADE_ID));
			end
		end,
		ScriptUse = function(v)
			local ts = TradeSkillMates[v[ROSTER_CHAR_NAME]];
			if( ts and #ts > 0 ) then
				return 1;
			end
			return nil;
		end,
	},
	class = {
		Label = L["Class"],
		Brush = function(v)
			if( iGuild.db.Column.class.Icon == true ) then
				return "|TInterface\\Addons\\iGuild\\Images\\"..v[ROSTER_CHAR_CLASS]..":14:14|t";
			end
			
			local option = iGuild.db.Column.class.Color;
			if( option == 2 ) then
				local c = RAID_CLASS_COLORS[v[ROSTER_CHAR_CLASS]];
				return ("|cff%02x%02x%02x%s|r"):format(c.r *255, c.g *255, c.b *255, v[ROSTER_CHAR_CLASS_LOCALIZED]);
			else
				return ("%s%s|r"):format(COLOR_GOLD, v[ROSTER_CHAR_CLASS_LOCALIZED]);
			end
		end,
	},
	exp = {
		Label = XP,
		Brush = function(v)
			return ("%s%d|r"):format(COLOR_GOLD, v[ROSTER_XP_MAX] /1000);
		end,
	},
};

local DisplayedColumns = {};
function iGuild:GetDisplayedColumns()
	tclear(DisplayedColumns);
	
	local insert;
	--for i = 1, #self.db.Display do
	for i = 1, #self.ConfigColumns do
		insert = true;
		--local canUse = self.Columns[self.db.Display[i]].CanUse;
		local canUse = self.Columns[self.ConfigColumns[i]].CanUse;
		
		if( type(canUse) == "boolean" and not canUse ) then
			insert = false;
		elseif( type(canUse) == "function" and not canUse() ) then
			insert = false;
		end
		
		if( insert ) then
			--table.insert(DisplayedColumns, self.db.Display[i]);
			table.insert(DisplayedColumns, self.ConfigColumns[i]);
		end
	end
	
end

----------------------
-- OnInitialize
----------------------

function iGuild:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("iGuildDB", self:CreateDB(), "Default").profile;

	-- dirty check if someone is using an old iGuild config, where this table-field was a table too
	if( type(self.db.Display) == "table" ) then
		self.db.Display = "level, name, zone, rank";
	end

	self:GetConfigColumns();
	self:GetDisplayedColumns();

	self:RegisterEvent("GUILD_MOTD", "RosterUpdate");
	self:RegisterEvent("PLAYER_GUILD_UPDATE", "RosterUpdate");
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "EnterWorld");
end

function iGuild:EnterWorld()
	if( IsInGuild() ) then
		self:RegisterEvent("GUILD_ROSTER_UPDATE", "RosterUpdate");
		self:RegisterEvent("GUILD_XP_UPDATE", "RosterUpdate");
		self:RegisterEvent("GUILD_TRADESKILL_UPDATE", "TradeSkillUpdate");
		
		GuildRoster();
		QueryGuildXP();
		QueryGuildRecipes();
		
		RosterTimer = self:ScheduleRepeatingTimer(GuildRoster, 15);
	else
		self.Feed.text = L["No guild"];
	end
end

----------------------
-- RosterUpdate
----------------------

function iGuild:RosterUpdate(event)
	if( IsInGuild() ) then
		if( not RosterTimer ) then
			self:EnterWorld();
		end
		
		local total, totalOn = GetNumGuildMembers();
		local feedText = ("%d/%d"):format(totalOn, total);
		local level = GetGuildLevel();
		local guild = GetGuildInfo("player");
		
		if( self.db.ShowGuildName and guild ) then
			feedText = ("%s%s:|r %s"):format(COLOR_GOLD, guild, feedText);
		end
		
		if( self.db.ShowGuildLevel ) then
			feedText = ("%s %s| |r%d"):format(feedText, COLOR_GOLD, GetGuildLevel());
		end
		
		if( self.db.ShowGuildXP ) then
			local currXP, nextUp = UnitGetGuildXP("player");
			feedText = ("%s (%d%%)"):format(feedText, level < MAX_GUILD_LEVEL and math.ceil(currXP / (currXP + nextUp) * 100) or 100);
		end
		
		self.Feed.text = feedText;		
		self:SetupGuildRoster();
		
		if( event == "GUILD_XP_UPDATE" ) then
			self:UnregisterEvent("GUILD_XP_UPDATE");
		end
	else
		if( RosterTimer ) then
			self:UnregisterEvent("GUILD_ROSTER_UPDATE");
			self:UnregisterEvent("GUILD_XP_UPDATE");
			self:UnregisterEvent("GUILD_TRADESKILL_UPDATE");
			self:CancelTimer(RosterTimer);
			RosterTimer = nil;
		end
		
		tclear(Roster);
		self:TradeSkillRemove();
		self.Feed.text = L["No guild"];
	end
	
	if( QTip:IsAcquired("iSuite"..AddonName) ) then
		self:UpdateTooltip();
	end
end

--------------------------
-- TradeskillUpdate
--------------------------

local TradeSkillsFetched;
local TradeSkillsUpdating;

function iGuild:TradeSkillUpdate()
	if( self.db.Column.tradeskills.Enable == false ) then
		return;
	end
	
	-- no tradeskill db, so we generate it :)
	if( not TradeSkillDB ) then
		TradeSkillDB = {};
		
		for i = 1, GetNumGuildTradeSkill() do
			local skillID, isCollapsed, iconTexture, headerName, _, _, _, _, _, _, _, _, _, _ = GetGuildTradeSkillInfo(i);
			
			if( headerName ) then
				TradeSkillDB[headerName] = {};
				TradeSkillDB[headerName][TRADE_ID] = skillID;
				TradeSkillDB[headerName][TRADE_ICON] = iconTexture;
				TradeSkillDB[headerName][TRADE_COLLAPSED] = isCollapsed;
			end
		end
	end
	
	if( TradeSkillsFetched or TradeSkillsUpdating ) then
		return;
	end
	
	TradeSkillsUpdating = 1;
	
	-- expand them all
	for _, v in pairs(TradeSkillDB) do
		ExpandGuildTradeSkillHeader(v[TRADE_ID]);
	end
	
	-- loop through all tradeskills and users
	tclear(TradeSkillMates);
	local showOffline = GetGuildRosterShowOffline();
	SetGuildRosterShowOffline(true);
	
	local currentTradeSkill;
	for i = 1, GetNumGuildTradeSkill() do
		local _, _, _, headerName, _, _, _, playerName, _, _, _, _, _, _ = GetGuildTradeSkillInfo(i);
		
		if( headerName ) then
			currentTradeSkill = headerName;
		elseif( playerName ) then
			if( not TradeSkillMates[playerName] ) then
				TradeSkillMates[playerName] = {};
			end
			
			table.insert(TradeSkillMates[playerName], currentTradeSkill);
		end
	end
	
	SetGuildRosterShowOffline(showOffline);
	
	-- collapse them again if collapsed before
	for _, v in pairs(TradeSkillDB) do		
		if( v[TRADE_COLLAPSED] ) then
			CollapseGuildTradeSkillHeader(v[TRADE_ID]);
		end
	end
	
	TradeSkillsFetched = 1;
	TradeSkillsUpdating = nil;
end

function iGuild:TradeSkillRemove()
	tclear(TradeSkillDB, 1);
	tclear(TradeSkillMates);
end

function iGuild:GetTradeSkill(name, index)
	if( index ) then
		return TradeSkillDB[name][index];
	end
	return TradeSkillDB[name];
end

--------------------------
-- SetupGuildRoster
--------------------------

local GuildIter = 1;
function iGuild:SetupGuildRoster()
	local total, totalOn = GetNumGuildMembers();
	
	tclear(Roster);
	
	-- little fix: iGuild couldn't show the roster when the ShowOffline option was enabled.
	-- the following if-statement prevents iGuild from always looping through all members, even if offliners aren't shown.
	if( not GetGuildRosterShowOffline() ) then
		total = totalOn;
	end
	
	for i = 1, total do
		local charName, guildRank, guildRankN, charLevel, charClassLoc, charZone, 
			guildNote, officerNote, isOnline, charStatus, charClass, acmPoints, acmRank, charMobile = GetGuildRosterInfo(i);
		local _, maxXP, _, _ = GetGuildRosterContribution(i);
		
		if( isOnline ) then
			Roster[GuildIter] = {};
			Roster[GuildIter][ROSTER_CHAR_NAME] = charName;
			Roster[GuildIter][ROSTER_CHAR_LEVEL] = charLevel;
			Roster[GuildIter][ROSTER_CHAR_CLASS] = charClass;
			Roster[GuildIter][ROSTER_CHAR_CLASS_LOCALIZED] = charClassLoc;
			Roster[GuildIter][ROSTER_CHAR_ZONE] = charZone or UNKNOWN;
			Roster[GuildIter][ROSTER_CHAR_STATUS] = charStatus;
			Roster[GuildIter][ROSTER_CHAR_MOBILE] = charMobile or false;
			Roster[GuildIter][ROSTER_ACM_POINTS] = acmPoints;
			Roster[GuildIter][ROSTER_ACM_RANK] = acmRank;
			Roster[GuildIter][ROSTER_GUILD_RANK] = guildRank;
			Roster[GuildIter][ROSTER_GUILD_RANK_NUM] = guildRankN;
			Roster[GuildIter][ROSTER_GUILD_NOTE] = guildNote or nil;
			Roster[GuildIter][ROSTER_OFFICER_NOTE] = officerNote or nil;
			Roster[GuildIter][ROSTER_XP_MAX] = maxXP;
			
			if( self.db.Column.tradeskills.Enable == true and TradeSkillMates[charName] ) then
				Roster[GuildIter][ROSTER_TRADESKILLS] = TradeSkillMates[charName];
			end
			
			GuildIter = GuildIter + 1;
		end
	end
	
	table.sort(Roster, self.Sort[self.db.Sort]);
	GuildIter = 1;
end

-----------------------
-- UpdateTooltip
-----------------------

local function LineClick(_, name, button)
	if( IsAltKeyDown() ) then
		InviteUnit(name);
	else
		SetItemRef(("player:%s"):format(name), ("|Hplayer:%s|h[%s]|h"):format(name, name), "LeftButton");
	end
end

function iGuild:UpdateTooltip()
	Tooltip:Clear();
	Tooltip:SetColumnLayout(#DisplayedColumns);
	
	local name, info;
	
	for y = 0, #Roster do
		local member;
		
		for x = 1, #DisplayedColumns do
			name = DisplayedColumns[x];
			info = self.Columns[name];
			
			if( x == 1 ) then
				if( y == 0) then
					Tooltip:AddHeader(" ");
				else
					Tooltip:AddLine(" ");
				end
			end
			
			if( y == 0 ) then
				if( self.db.Column[name].ShowLabel ) then
					Tooltip:SetCell(y +1, x, info.Label, nil, self.db.Column[name].Align);
				end
				
			else
				member = Roster[y];
				Tooltip:SetCell(y +1, x, info.Brush(member), nil, self.db.Column[name].Align);
				
				if( info.Script and info.ScriptUse(member) ) then
					Tooltip:SetCellScript(y +1, x, "OnMouseDown", info.Script, member);
				end
			end
		end
		
		if( member ) then
			Tooltip:SetLineScript(y +1, "OnMouseDown", LineClick, member[ROSTER_CHAR_NAME]);
		end
	end
end